#version 330
#extension GL_ARB_separate_shader_objects : enable





layout(location = 0) in vec3 _vin_Position;
layout(location = 1) in vec3 _vin_Normal;
layout(location = 2) in vec2 _vin_TexCoord;
layout(location = 3) in vec4 _vin_Color;

struct SVertexInput
{
	vec3 Position;
	vec3 Normal;
	vec2 TexCoord;
	vec4 Color;
	int VertexIndex;
	uint InstanceId;
};

layout(location = 0) smooth out vec3 _vout_fragNormal;
layout(location = 1) smooth out vec2 _vout_fragTexCoord;
layout(location = 2) smooth out vec4 _vout_fragWorldPosition;
layout(location = 3) smooth out vec4 _vout_fragColor;

struct SVertexOutput
{
	vec3 fragNormal;
	vec2 fragTexCoord;
	vec4 fragWorldPosition;
	vec4 fragColor;
	vec4 VertexPosition;
};

layout( std140) uniform USceneInfo
{
	mat4 ViewProj;
	mat4 View;
	mat4 Proj;
	mat4 InvView;
	mat4 InvViewProj;
	mat4 UserToLHZ;
	mat4 LHZToUser;
	mat4 BillboardingView;
	mat4 PackNormalView;
	mat4 UnpackNormalView;
	vec4 SideVector;
	vec4 DepthVector;
	vec2 ZBufferLimits;
	vec2 ViewportSize;
	float Handedness;
} SceneInfo;

layout( std140) uniform UMeshVertexConstant
{
	mat4 ModelMatrix;
	mat4 NormalMatrix;
} MeshVertexConstant;



void VertexMain( in SVertexInput  vInput,  out SVertexOutput  vOutput  )
{
	mat4 mvp =  (( SceneInfo.ViewProj ) * ( MeshVertexConstant.ModelMatrix )) ;
    vOutput.VertexPosition =  ((mvp) * (vec4(vInput.Position, 1.0))) ;
    vOutput.fragNormal =  (( MeshVertexConstant.NormalMatrix ) * (vec4(vInput.Normal, 0))) .xyz;
    vOutput.fragTexCoord = vInput.TexCoord;
	vOutput.fragWorldPosition = vOutput.VertexPosition;

	vOutput.fragColor = vInput.Color;

}


void main()
{
	SVertexInput vInput;
	SVertexOutput vOutput;

	vInput.Position = _vin_Position;
	vInput.Normal = _vin_Normal;
	vInput.TexCoord = _vin_TexCoord;
	vInput.Color = _vin_Color;
	vInput.VertexIndex = gl_VertexID;
	vInput.InstanceId = uint(gl_InstanceID);
	vOutput.fragNormal = vec3(0, 0, 0);
	vOutput.fragTexCoord = vec2(0, 0);
	vOutput.fragWorldPosition = vec4(0, 0, 0, 0);
	vOutput.fragColor = vec4(0, 0, 0, 0);
	VertexMain(vInput, vOutput);
	_vout_fragNormal = vOutput.fragNormal;
	_vout_fragTexCoord = vOutput.fragTexCoord;
	_vout_fragWorldPosition = vOutput.fragWorldPosition;
	_vout_fragColor = vOutput.fragColor;
	gl_Position = vOutput.VertexPosition;
}
